'''
==========
Date Aug 27, 2021
OAMP 
==========
U0 -> G0 -> V0 -> F0 -> U1 -> G1 -> amp -> V1 -> F1
Input niters
Output 
U1, U2,... U_{n+1}
V0, V1, V2, .. V_{n}
'''

import numpy as np
np.seterr(all = "raise")
from numpy.linalg import matrix_power as matprod

from pca import RecPcaPack, get_gaussian_pcapack_from_pcapack

from se import *
from empbayes import NonparEB


def processUV(U, V, niters, rank, tu, tv):
    '''output [U_1,...U_t] and [V_0,...,V_t]
    '''
    m = U.shape[0]
    n = V.shape[0]
    U = U.reshape((m, rank, -1), order = 'F')[:,:,1:(tu+1)]
    V = V.reshape((n, rank, -1), order = 'F')[:,:,:(tv+1)]
    return U,V


def sriamp(pcapack: RecPcaPack, udenoiser = NonparEB(), vdenoiser = NonparEB(), \
        niters = 5, SeModel = SRISE):
    
    return riamp(pcapack, udenoiser, vdenoiser, \
        niters = niters, SeModel = SeModel)

def rriamp(pcapack: RecPcaPack, udenoiser = NonparEB(), vdenoiser = NonparEB(), \
        niters = 5, SeModel = RRISE):
    
    return riamp(pcapack, udenoiser, vdenoiser, \
        niters = niters, SeModel = RRISE)

def riamp(pcapack: RecPcaPack, udenoiser = NonparEB(), vdenoiser = NonparEB(), \
        niters = 5, SeModel = RISE):
    Y = pcapack.Y 
    m, n = Y.shape 
    aspect_ratio = m/n 
    K = len(pcapack.thetas)

    # Init StateEvolution
    SE = SeModel(K, niters, aspect_ratio, \
        pcapack.thetas, pcapack.uthetas, pcapack.vthetas, \
            pcapack.kappas, pcapack.zetas, pcapack.dzetas)
    
    # Init Iterates 
    F = pcapack.F # (f_0, ..., f_{T})
    G = pcapack.G # (g_1, ..., g_T)

    U = np.zeros((m, (niters+1)*K))
    U[:,:K] = U[:,K:2*K] = F/pcapack.uthetas
    V = np.zeros((n, (niters+1)*K))
    V[:,:K] = G/pcapack.vthetas


    # Init state 
    Mu = np.diag(pcapack.uinnerprods)
    Sigma = np.identity(K) - Mu**2
    Nu = np.diag(pcapack.vinnerprods)
    Omega = np.identity(K) - Nu**2
    
    udenoiser.fit(F, Mu, Sigma, figname = str(0))
    
    for t in range(1, niters+1):
        print("At iter", t)
        try: 
            # Denoise Gt for vt
            vdenoiser.fit(G, Nu, Omega, figname = str(t))
            vt = vdenoiser.denoise(G, Nu, Omega) # Vt = v_t(G1,...,Gt)
        except Exception as e:
            print("Stopped due to an error to denoise vt at t={}:\n{}".format(t,e))
            return processUV(U,V,niters, K, t, t-1)
        
        V[:,t*K:(t+1)*K] = vt
    
        try: 
            # Update Internal State 
            Gamma = (V.T @ V / n)[:(t+1)*K, :(t+1)*K]
            SE.updateGamma(t, Gamma)
            gradv = vdenoiser.ddenoise(G, Nu, Omega)
            SE.updatePsi(t, gradv.mean(axis = 0))

            # Next Iteration 
            a = SE.geta(t)
            ft = Y.dot(vt) - U.dot(a.T) 
            F = np.hstack((F, ft)) # (F0,...,Ft)

            # Update Mu and Sigma
            Sigma = SE.getSigma(t)
            mut = SE.getMut(t)
            Mu = np.vstack((Mu, mut))

            # Denoise ft for utp1
            udenoiser.fit(F, Mu, Sigma, figname = str(t))
            utp1 = udenoiser.denoise(F, Mu, Sigma)
        except Exception as e:
            print("Stopped due to an error to denoise utp1 at t={}:\n{}".format(t,e))
            return processUV(U,V,niters, K, t, t)
        
        if t == niters:
            U = np.hstack((U, utp1))
            return processUV(U,V,niters, K, t+1, t)
        
        U[:,(t+1)*K:(t+2)*K] = utp1

        
        try: 
            # U = [u0, ..., u_{t+1}]
            # Update Internal State
            Delta = (U.T @ U / m)[:(t+2)*K, :(t+2)*K]
            SE.updateDelta(t+1, Delta)
            gradu = udenoiser.ddenoise(F, Mu, Sigma)
            SE.updatePhi(t+1, gradu.mean(axis = 0))

            # Debiasing coef B
            b = SE.getb(t+1)

            # Get gtp1 from utp1
            gtp1 = Y.T.dot(utp1) - V.dot(b.T) * aspect_ratio
            G = np.hstack((G, gtp1)) # (G1, ... ,Gt+1)

            # Update Nu and Omega
            Omega = aspect_ratio * SE.getOmega(t+1)
            nutp1 = SE.getNut(t+1)
            Nu = np.vstack((Nu, nutp1))
        
        except Exception as e:
            print("Stopped due to an error to create states for gtp1 at t={}:\n{}".format(t,e))
            return processUV(U,V, niters, K, t+1, t)



    
def riunamp(pcapack: RecPcaPack, udenoiser = NonparEB(), vdenoiser = NonparEB(), \
        niters = 5, SeModel = RISE):
    
    Y = pcapack.Y 
    m, n = Y.shape 
    aspect_ratio = m/n 
    K = len(pcapack.thetas)

    # Init StateEvolution
    SE = SeModel(K, niters, aspect_ratio, \
        pcapack.thetas, pcapack.uthetas, pcapack.vthetas, \
            pcapack.kappas, pcapack.zetas, pcapack.dzetas)
    
    # Init Iterates 
    f = pcapack.F # (f_0, ..., f_{T})
    gtp1 = pcapack.G # (g_1, ..., g_T)

    U = np.zeros((m, (niters+1)*K))
    U[:,:K] = U[:,K:2*K] = f/pcapack.uthetas
    V = np.zeros((n, (niters+1)*K))
    V[:,:K] = gtp1/pcapack.vthetas

    # Init state
    mut = np.diag(pcapack.uinnerprods)
    sigmat = np.identity(K) - mut**2
    nutp1 = np.diag(pcapack.vinnerprods)
    omegatp1 = np.identity(K) - nutp1**2
    
    udenoiser.fit(f, mut, sigmat, figname = str(0))

    for t in range(1, niters+1):
        try: 
            # Denoise Gt for vt
            vdenoiser.fit(gtp1, nutp1, omegatp1, figname = str(t))
            vt = vdenoiser.denoise(gtp1, nutp1, omegatp1) # Vt = v_t(Gt)
        except Exception as e:
            print("Stopped due to an error to denoise vt at t={}:\n{}".format(t,e))
            return processUV(U,V,niters, K, t, t-1)
        
        V[:,t*K:(t+1)*K] = vt
    
        try: 
            # Update Internal State 
            Gamma = (V.T @ V / n)[:(t+1)*K, :(t+1)*K]
            SE.updateGamma(t, Gamma)
            gradv = vdenoiser.ddenoise(gtp1, nutp1, omegatp1)
            psiRow = np.hstack((np.zeros(shape = (K, K*(t-1))), \
                gradv.mean(axis = 0) ))
            SE.updatePsi(t, psiRow)

            # Next Iteration 
            a = SE.geta(t)
            ft = Y.dot(vt) - U.dot(a.T) 

            # Update Mu and Sigma
            Sigma = SE.getSigma(t)
            sigmat = Sigma[-K:,-K:]
            mut = SE.getMut(t)

            # Denoise ft for utp1
            udenoiser.fit(ft, mut, sigmat, figname = str(t))
            utp1 = udenoiser.denoise(ft, mut, sigmat)
        
        except Exception as e:
            print("Stopped due to an error to denoise utp1 at t={}:\n{}".format(t,e))
            return processUV(U,V,niters, K, t, t)
        
        if t == niters:
            U = np.hstack((U, utp1))
            return processUV(U,V,niters, K, t+1, t)
        
        U[:,(t+1)*K:(t+2)*K] = utp1

        
        try: 
            # U = [u0, ..., u_{t+1}]
            # Update Internal State
            Delta = (U.T @ U / m)[:(t+2)*K, :(t+2)*K]
            SE.updateDelta(t+1, Delta)
            gradu = udenoiser.ddenoise(ft, mut, sigmat)
            phiRow = np.hstack((np.zeros(shape = (K, K*t)), \
                gradu.mean(axis = 0)))
            SE.updatePhi(t+1, phiRow)

            # Debiasing coef B
            b = SE.getb(t+1)

            # Get gtp1 from utp1
            gtp1 = Y.T.dot(utp1) - V.dot(b.T) * aspect_ratio

            # Update Nu and Omega
            omegatp1 = aspect_ratio * SE.getOmega(t+1)[-K:,-K:]
            nutp1 = SE.getNut(t+1)
        
        except Exception as e:
            print("Stopped due to an error to create states for gtp1 at t={}:\n{}".format(t,e))
            return processUV(U,V, niters, K, t+1, t)




    
def gamp(pcapack: RecPcaPack, udenoiser = NonparEB(), vdenoiser = NonparEB(), \
        niters = 5, SeModel = RISE):
    
    Y = pcapack.Y 
    m, n = Y.shape 
    aspect_ratio = m/n 
    K = len(pcapack.thetas)

    # Init StateEvolution
    SE = GSE(K, niters, aspect_ratio, \
        pcapack.thetas, pcapack.uthetas, pcapack.vthetas)
    
    # Init Iterates 
    f = pcapack.F # (f_0, ..., f_{T})
    gtp1 = pcapack.G # (g_1, ..., g_T)

    U = np.zeros((m, (niters+1)*K))
    U[:,:K] = U[:,K:2*K] = f/pcapack.uthetas
    V = np.zeros((n, (niters+1)*K))
    V[:,:K] = gtp1/pcapack.vthetas

    # Init state
    mut = np.diag(pcapack.uinnerprods)
    sigmat = np.identity(K) - mut**2
    nutp1 = np.diag(pcapack.vinnerprods)
    omegatp1 = np.identity(K) - nutp1**2
    
    udenoiser.fit(f, mut, sigmat, figname = str(1))

    for t in range(1, niters+1):
        try: 
            # Denoise Gt for vt
            vdenoiser.fit(gtp1, nutp1, omegatp1, figname = str(t))
            vt = vdenoiser.denoise(gtp1, nutp1, omegatp1) # Vt = v_t(Gt)
        except Exception as e:
            print("Stopped due to an error to denoise vt at t={}:\n{}".format(t,e))
            return processUV(U,V,niters, K, t, t-1)
        
        V[:,t*K:(t+1)*K] = vt
    
        try: 
            # Update Internal State 
            Gamma = (V.T @ V / n)[:(t+1)*K, :(t+1)*K]
            SE.updateGamma(t, Gamma)
            gradv = vdenoiser.ddenoise(gtp1, nutp1, omegatp1)
            psiRow = np.hstack((np.zeros(shape = (K, K*(t-1))), \
                gradv.mean(axis = 0) ))
            SE.updatePsi(t, psiRow)

            # Next Iteration 
            a = SE.geta(t)
            ft = Y.dot(vt) - U.dot(a.T) 

            # Update Mu and Sigma
            Sigma = SE.getSigma(t)
            sigmat = Sigma[-K:,-K:]
            mut = SE.getMut(t)

            # Denoise ft for utp1
            udenoiser.fit(ft, mut, sigmat, figname = str(t + 1))
            utp1 = udenoiser.denoise(ft, mut, sigmat)
        
        except Exception as e:
            print("Stopped due to an error to denoise utp1 at t={}:\n{}".format(t,e))
            return processUV(U,V,niters, K, t, t)
        
        if t == niters:
            U = np.hstack((U, utp1))
            return processUV(U,V,niters, K, t+1, t)
        
        U[:,(t+1)*K:(t+2)*K] = utp1

        
        try: 
            # U = [u0, ..., u_{t+1}]
            # Update Internal State
            Delta = (U.T @ U / m)[:(t+2)*K, :(t+2)*K]
            SE.updateDelta(t+1, Delta)
            gradu = udenoiser.ddenoise(ft, mut, sigmat)
            phiRow = np.hstack((np.zeros(shape = (K, K*t)), \
                gradu.mean(axis = 0)))
            SE.updatePhi(t+1, phiRow)

            # Debiasing coef B
            b = SE.getb(t+1)

            # Get gtp1 from utp1
            gtp1 = Y.T.dot(utp1) - V.dot(b.T) * aspect_ratio

            # Update Nu and Omega
            omegatp1 = aspect_ratio * SE.getOmega(t+1)[-K:,-K:]
            nutp1 = SE.getNut(t+1)
        
        except Exception as e:
            print("Stopped due to an error to create states for gtp1 at t={}:\n{}".format(t,e))
            return processUV(U,V, niters, K, t+1, t)

def ggamp(opcapack: RecPcaPack, udenoiser = NonparEB(), vdenoiser = NonparEB(), \
        niters = 5, SeModel = RISE):
    '''
    AMP assuming the noise matrix is Gaussian 

    '''
    pcapack = get_gaussian_pcapack_from_pcapack(opcapack)
    Y = pcapack.Y 
    m, n = Y.shape 
    aspect_ratio = m/n 
    K = len(pcapack.thetas)

    # Init StateEvolution

    SE = GSE(K, niters, aspect_ratio, \
        pcapack.thetas, pcapack.uthetas, pcapack.vthetas)
    
    # Init Iterates 
    f = pcapack.F # (f_0, ..., f_{T})
    gtp1 = pcapack.G # (g_1, ..., g_T)

    U = np.zeros((m, (niters+1)*K))
    U[:,:K] = U[:,K:2*K] = f/pcapack.uthetas
    V = np.zeros((n, (niters+1)*K))
    V[:,:K] = gtp1/pcapack.vthetas

    # Init state
    mut = np.diag(pcapack.uinnerprods)
    sigmat = np.identity(K) - mut**2
    nutp1 = np.diag(pcapack.vinnerprods)
    omegatp1 = np.identity(K) - nutp1**2
    
    udenoiser.fit(f, mut, sigmat, figname = str(1))

    for t in range(1, niters+1):
        try: 
            # Denoise Gt for vt
            vdenoiser.fit(gtp1, nutp1, omegatp1, figname = str(t))
            vt = vdenoiser.denoise(gtp1, nutp1, omegatp1) # Vt = v_t(Gt)
        except Exception as e:
            print("Stopped due to an error to denoise vt at t={}:\n{}".format(t,e))
            return processUV(U,V,niters, K, t, t-1)
        
        V[:,t*K:(t+1)*K] = vt
    
        try: 
            # Update Internal State 
            Gamma = (V.T @ V / n)[:(t+1)*K, :(t+1)*K]
            SE.updateGamma(t, Gamma)
            gradv = vdenoiser.ddenoise(gtp1, nutp1, omegatp1)
            psiRow = np.hstack((np.zeros(shape = (K, K*(t-1))), \
                gradv.mean(axis = 0) ))
            SE.updatePsi(t, psiRow)

            # Next Iteration 
            a = SE.geta(t)
            ft = Y.dot(vt) - U.dot(a.T) 

            # Update Mu and Sigma
            Sigma = SE.getSigma(t)
            sigmat = Sigma[-K:,-K:]
            mut = SE.getMut(t)

            # Denoise ft for utp1
            udenoiser.fit(ft, mut, sigmat, figname = str(t + 1))
            utp1 = udenoiser.denoise(ft, mut, sigmat)
        
        except Exception as e:
            print("Stopped due to an error to denoise utp1 at t={}:\n{}".format(t,e))
            return processUV(U,V,niters, K, t, t)
        
        if t == niters:
            U = np.hstack((U, utp1))
            return processUV(U,V,niters, K, t+1, t)
        
        U[:,(t+1)*K:(t+2)*K] = utp1

        
        try: 
            # U = [u0, ..., u_{t+1}]
            # Update Internal State
            Delta = (U.T @ U / m)[:(t+2)*K, :(t+2)*K]
            SE.updateDelta(t+1, Delta)
            gradu = udenoiser.ddenoise(ft, mut, sigmat)
            phiRow = np.hstack((np.zeros(shape = (K, K*t)), \
                gradu.mean(axis = 0)))
            SE.updatePhi(t+1, phiRow)

            # Debiasing coef B
            b = SE.getb(t+1)

            # Get gtp1 from utp1
            gtp1 = Y.T.dot(utp1) - V.dot(b.T) * aspect_ratio

            # Update Nu and Omega
            omegatp1 = aspect_ratio * SE.getOmega(t+1)[-K:,-K:]
            nutp1 = SE.getNut(t+1)
        
        except Exception as e:
            print("Stopped due to an error to create states for gtp1 at t={}:\n{}".format(t,e))
            return processUV(U,V, niters, K, t+1, t)

AMPALGO = {
    "riamp": riamp,
    "riunamp": riunamp,
    "gamp": gamp,
    "ggamp": ggamp,
    "sriamp": sriamp,
    "rriamp": rriamp
}


