
from numpy.core.defchararray import mod
from numpy.lib.function_base import vectorize


'''
==========

Ref:
The Singular Values and Vectors of Low Rank Perturbations of Large Rectangular 
Random Matrices 
Note:
Our matrix is (m,n), theirs is (n,m)
==========
'''

from abc import ABC, abstractmethod

import numpy as np
from numpy.polynomial import polynomial

## --- get kappas --- ##

def _get_moments_from_empirical_measure(eigs, k):
    m = []
    for j in range(k+1):
        m.append(np.mean(eigs**(2*j)))
    return np.array(m)

# Compute first K rectangular free cumulants from moments
#
# Input: m[0] = 1 and m[k] = mu_{2k} for k = 1,...,K
# Output: kappa[0] = 0 and kappa[k] = kappa_{2k} for k = 1,...,K
def _get_cumulants_from_moments(m,gamma,K): # gamma is the aspect ratio
    M = np.copy(m)
    M[0] = 0
    A = polynomial.polyadd(gamma*M,[1])
    B = polynomial.polyadd(M,[1])
    S = polynomial.polymul([0,1],polynomial.polymul(A,B))
    kappa = [0,M[1]]
    for k in range(2,K+1):
        P = [0]
        for j in range(1,k):
            P = polynomial.polyadd(P,kappa[j]*polynomial.polypow(S,j))
        kappa.append(M[k]-P[k])
    return np.array(kappa)

## --- Cauchy Transform --- ##

def _cauchy_transform(z, mu, n_samples, n_features):
    prod = np.mean(z / (z**2 - mu[:n_samples]**2))
    prod = np.mean(z/(z**2 - mu[:n_features]**2))*prod
    return prod 

def _vec_cauchy_transform(z, mu, n_samples, n_features):
    '''vectorize over the first argument'''
    inner = lambda x: _cauchy_transform(x, mu, n_samples, n_features)
    return np.vectorize(inner)(z)

def _d_cauchy_transform(z, mu, n_samples, n_features):
    a = np.mean(z / (z**2 - mu[:n_samples]**2))
    b = np.mean(z/(z**2 - mu[:n_features]**2))
    da = -np.mean((z**2 + mu[:n_samples]**2)/ (z**2 - mu[:n_samples]**2)**2)
    db = -np.mean( (z**2 + mu[:n_features]**2)/ (z**2 - mu[:n_features]**2)**2)
    return da*b + db*a 

def _vec_d_cauchy_transform(z, mu, n_samples, n_features):
    '''vectorize over the first argument'''
    inner = lambda x:_d_cauchy_transform(x, mu, n_samples, n_features)
    return np.vectorize(inner)(z)

def _phase_transition_threshold(mu, n_samples, n_features, tol = 0.01):
    supp_max_plus = mu.max() + tol 
    mu = np.pad(mu, (0,(max(n_features, n_samples) - len(mu))), constant_values = (0,0))
    ct = _cauchy_transform(supp_max_plus, mu, n_samples, n_features)
    return 1/np.sqrt(ct)


## --- Cosmetic --- ##

class NoiseGenFactory:
    """ The factory class for all defined noises"""

    registry = {}
    """ Internal registry for available noises """

    @classmethod
    def register(cls, name: str):
        """ Class method to register Noise class to the internal registry.
        Args:
            name (str): The name of the noise.
        Returns:
            The specific noise class itself.
        """

        def inner_wrapper(wrapped_class):
            if name in cls.registry:
                print('%s noise already exists. Will replace it' % name)
            cls.registry[name] = wrapped_class
            return wrapped_class

        return inner_wrapper

class NoiseChef():

    def __init__(self, mu, m, n):
        self.m, self.n = m, n
        self.mu = np.pad(mu, (0,(max(m, n) - len(mu))), constant_values = (0,0))
        self.thetas = None
        self.singvals = None 
        self.dCauchy = None

    def estimateThetas(self, singvals, tol = 0.01):
        self.singvals = singvals
        supp_max = self.mu.max()
        if singvals.min() <= (supp_max+tol):
            raise ValueError("signal must seperate from singval supp of noise matrix")
        self.thetas = np.sqrt(1/_vec_cauchy_transform(singvals, self.mu, self.m, self.n))   
        self.dCauchy = _vec_d_cauchy_transform(singvals, self.mu, self.m, self.n)
        return self.thetas 
    
    def checkInfo(self, singvals):
        if self.singvals is None:
            if singvals is not None:
                _ = self.estimateTheta(singvals)
            else: 
                raise RuntimeError("Without singvals we can not estimate.")
        if (self.singvals is not None) and ( np.array_equal(self.singvals, singvals)):
            raise Warning("Singvals are not consisitent. Update with the latest")

    def estimatePhaseTransition(self):
        return _phase_transition_threshold(self.mu, self.m, self.n)

    @staticmethod
    def _getphi(singval, mu, ndim):
        return np.mean(singval/(singval**2 - mu[:ndim]**2))
    
    @staticmethod 
    def _vec_getphi(singvals, mu, ndim):
        inner = lambda x: NoiseChef._getphi(x, mu, ndim)
        return vectorize(inner)(singvals)

    def estimateUInnerprods(self, singvals = None):
        self.checkInfo(singvals)
        phi = NoiseChef._vec_getphi(self.singvals, self.mu, self.m)
        sample_aligns = np.sqrt(-(2 * phi/self.dCauchy))/self.thetas
        return sample_aligns
    
    def estimateVInnerprods(self, singvals = None):
        self.checkInfo(singvals)
        phi = NoiseChef._vec_getphi(self.singvals, self.mu, self.n)
        feature_aligns = np.sqrt(-(2 * phi/self.dCauchy))/self.thetas
        return feature_aligns

    def getCumulants(self, J):
        moments = _get_moments_from_empirical_measure(self.mu[:min(self.n, self.m)], J)
        gamma = self.m/self.n
        kappas =  _get_cumulants_from_moments(moments, gamma, J)   
        # print("moments")
        # print(moments)
        # print("kappas") 
        # print(kappas)
        return kappas
        # mp = np.zeros(shape = kappas.shape)
        # mp[1] = 1.0
        # return mp
    



class _NoiseGenerator(ABC):

    def __init__(self, m, n, seed):
        self.m = m
        self.n = n
        self.rng = np.random.default_rng(seed)
        self.W = self.generateW()
    
    
    def getMu(self, minmn):
        mu = self.getmu(minmn)
        return mu/np.sqrt((mu**2).mean())
    
    @abstractmethod
    def getmu(self, minmn):
        pass 

    def generateW(self):
        mu = self.getMu(min(self.n,self.m))
        D = np.zeros((self.m,self.n))
        np.fill_diagonal(D,mu)
        W = self.rng.normal(size=(self.m,self.m))
        O = np.linalg.qr(W)[0]
        W = self.rng.normal(size=(self.n,self.n))
        Q = np.linalg.qr(W)[0]
        W = np.dot(np.dot(O,D),Q)
        return W



@NoiseGenFactory.register("MP")
class MP():
    name = "MP"
    def __init__(self, m, n, rngseed):
        self.m = m
        self.n = n
        self.rng = np.random.default_rng(rngseed)
        self.W = self.generateW()
    
    def generateW(self):
        return self.rng.normal(size=(self.m,self.n))/np.sqrt(self.n)

@NoiseGenFactory.register("Gamma")
class Gamma(_NoiseGenerator):
    name = "Gamma"
    def __init__(self, m, n, rngseed):
        _NoiseGenerator.__init__(self, m, n, rngseed)

    def getmu(self, minmn):
        return self.rng.gamma(1,2,minmn)

# @NoiseGenFactory.register("Beta")
# class Beta(_NoiseGenerator):
#     name = "Beta"
#     def __init__(self, m, n):
#         _NoiseGenerator.__init__(self, m, n)

#     def getmu(self, minmn):
#         return self.rng.beta(1,2,minmn)

@NoiseGenFactory.register("Uniform")
class Uniform(_NoiseGenerator):
    name = "Uniform"
    def __init__(self, m, n, rngseed):
        _NoiseGenerator.__init__(self, m, n, rngseed)

    def getmu(self, minmn):
        return self.rng.uniform(1,2,minmn)

@NoiseGenFactory.register("Point")
class Point(_NoiseGenerator):
    name = "Point"
    def __init__(self, m, n, rngseed):
        _NoiseGenerator.__init__(self, m, n, rngseed)
    
    def getmu(self, minmn):
        return np.full((minmn, ), 1)

@NoiseGenFactory.register("Beta")
class Point(_NoiseGenerator):
    name = "Beta"
    def __init__(self, m, n, rngseed):
        _NoiseGenerator.__init__(self, m, n, rngseed)
    
    def getmu(self, minmn):
        return self.rng.beta(3,1,size=minmn)


    




    










