'''
==========
This module support infer 
1. initial estimates of the spikes
2. empirical estiamtes of the greeks of the noise
==========
'''

from collections import namedtuple

import numpy as np

from noise import NoiseChef


RecPcaPack = namedtuple("RecPcaPack",["Y", "F", "G", \
    "aspect_ratio", "singvals", "thetas", \
        "kappas", "zetas", "dzetas", \
            "uthetas", "uinnerprods", \
                "vthetas", "vinnerprods"])

def gaussianDtransform(z, c):
    '''
    c: aspect ratio 
    '''
    return (z**2 - (c+1) - np.sqrt((z**2 - (c+1))**2 - 4 * c)) / 2 /c 

def get_gaussian_pcapack_from_pcapack(opcapack: RecPcaPack):

    singvals = opcapack.singvals
    aspect_ratio = opcapack.aspect_ratio 

    # Phase transition threshold for Gaussian model
    print("singvals", singvals)
    assert singvals.min() > 1 + np.sqrt(aspect_ratio)

    # theta = 1/sqrt(singvals)
    thetas = 1/np.sqrt(gaussianDtransform(singvals, aspect_ratio))
    ualigns = np.sqrt(thetas**4 - aspect_ratio) / np.sqrt(thetas**2 + aspect_ratio) / thetas 
    valigns = np.sqrt(thetas**4 - aspect_ratio) / np.sqrt(thetas**2 + 1) / thetas 

    uthetas = np.sqrt(aspect_ratio) * singvals / (1 + aspect_ratio/thetas**2)
    vthetas = singvals / (1 + 1/thetas**2) / np.sqrt(aspect_ratio)

    
    gpcapack = RecPcaPack(Y = opcapack.Y, F = opcapack.F, G = opcapack.G, \
        aspect_ratio = opcapack.aspect_ratio, singvals=opcapack.singvals, thetas=thetas,\
            kappas = 0, zetas = 0, dzetas = 0, \
                uthetas = uthetas, uinnerprods=ualigns,
                vthetas = vthetas, vinnerprods=valigns
                )
    
    return gpcapack

        






# R(1/thetasq) = \sum_{i=1}^\infty kappa_{2j}/theta^{2j}
# zeta0 = \sum_{i=0}^\infty kappa_{2j}/theta^{2j}
def get_zeta0(R, kappa0):
    return R - kappa0

def get_dR(deltaPCA, R, gamma, thetasq):
    '''
    See FZ paper Eq. C.32
    dR = \sum_{k=0} (k+1) kappa_2{j+1} z^k
    '''
    TR = (1 + R)*(1 + gamma*R)
    x = 1/thetasq
    dTR = 2*gamma*R + 1 + gamma
    dR = (TR - deltaPCA*(1 + gamma * R)) / (x * dTR)
    return dR 

def get_dzeta0(R, thetasq, deltaPCA, gamma, zetas):
    '''
    dR = \sum_{j=0}^\infty \kappa_2{j+1} (j+1) z^j at z = 1/thetasq
    dzeta_{0}(z) = \sum_{j=0}^\infty (j+1) kappa_2{j} z^j
    In fact, dR is dzeta_{1}
    '''
    dR = get_dR(deltaPCA, R, gamma, thetasq)
    return dR/thetasq + zetas[0]
    # (dR + zetas[0])/thetasq  # NOTE: changed the definition of dR


# zetas[j] = zetas_{2j}
def get_zetas_from_kappas(zeta0, thetasq, kappas):
    zetas = [zeta0]
    prev = zeta0
    for i in range(len(kappas)):
        prev = (prev - kappas[i])*thetasq
        zetas.append(prev)
    return np.array(zetas)


def get_dzetas_from_zetas(dzeta0, thetasq, zetas):
    dzetas = [dzeta0]
    prev = dzeta0
    for i in range(len(zetas)):
        prev = (prev - zetas[i])*thetasq
        dzetas.append(prev)
    return np.array(dzetas)

def solve_thetauv(lmda, theta, gamma):
    '''
    Get R(1/thetasq), theta_u and theta_v, see note Mar 25.
    '''
    a = gamma
    b = (gamma + 1)
    c = 1 - lmda**2/theta**2
    R = (-b + np.sqrt(b**2 - 4*a*c))/ (2*a)
    theta_v = lmda / (np.sqrt(gamma)*(1+R))
    theta_u = lmda*np.sqrt(gamma) / (1+gamma * R)
    return R, theta_u, theta_v

def get_greeks(gamma, thetas, R, uinnerprods, kappas):
    
    thetasq = thetas**2
    
    zeta0 = get_zeta0(R, kappas[0])
    zetas = get_zetas_from_kappas(zeta0, thetasq, kappas)
    deltaPCAs = uinnerprods**2
    dzeta0 = get_dzeta0(R, thetasq, deltaPCAs, gamma, zetas)
    dzetas = get_dzetas_from_zetas(dzeta0, thetasq, zetas)
    return kappas, zetas, dzetas    


def get_pcapack(Y, rank, niters):
    '''
    1. Infer the sample PCs as init
    2. get kappas from matrix residual
    3. get other greeks
    '''
    m, n = Y.shape 
    assert m <= n
    gamma = m/n
    K = rank; J = 2*niters + 2

    F, S, Gh = np.linalg.svd(Y)
    
    fsvd = F[:,:K] * np.sqrt(m)
    gsvd = np.transpose(Gh[:K,:] * np.sqrt(n))
    
    singvals = S[:K]
    print("sigvals", singvals)
    
    noiseChef = NoiseChef(S[K:], m, n)
    print("Phase Transition Threshould", noiseChef.estimatePhaseTransition())

    thetas = noiseChef.estimateThetas(singvals)
    print("Estimated Thetas:", thetas)

    R, uthetas, vthetas = solve_thetauv(singvals, thetas, gamma)

    uinnerprods = noiseChef.estimateUInnerprods()
    vinnerprods = noiseChef.estimateVInnerprods()

    kappas = noiseChef.getCumulants(J)

    # Redefine get greeks
    kappas, zetas, dzetas = get_greeks(gamma, thetas, R, uinnerprods, kappas)

    pcapack = RecPcaPack(Y = Y, F = fsvd, G = gsvd, \
        aspect_ratio = gamma, singvals = singvals, thetas = thetas, \
            kappas = kappas, zetas = zetas, dzetas = dzetas, \
            uthetas = uthetas, uinnerprods=uinnerprods,\
            vthetas = vthetas, vinnerprods=vinnerprods)
    
    return pcapack 


