import os 

import pandas as pd 
import numpy as np
import matplotlib.pyplot as plt

from utils import mkdir 
from vis import getPos, set_box_color

MAXITER = 5

figTitles = {
    "MP": "MP",
    "Beta": "Beta",
    "Point": "Point",
    "Uniform": "Uniform"
}

boxcolors = {
    "sriamp": "red",
    "riunamp": "#069AF3",
    "ggamp": "green", 
}

# 
# def read_algo_file(fname, maxiter = MAXITER): # MODIFIED
#     df = pd.read_csv(fname, header = None)
#     # df = df.iloc[:,:MAXITER]
#     df.columns = [str(i) for i in range(df.shape[1])]
#     df["trial"] = np.arange(len(df))
#     return df.tail(50)

# This version substitutes U_0 by F_0
def read_algo_file(fpattern, algoname, maxiter = MAXITER):
    fname = fpattern.replace("PLACEHOLDER", str(algoname))
    df = pd.read_csv(fname, header = None)
    df = df.iloc[:,:(MAXITER + 1)]
    df.columns = [str(i) for i in range(df.shape[1])]
    df["trial"] = np.arange(len(df))
    # substitue U_0 by F_0
    fname = fpattern.replace("PLACEHOLDER", "pca")
    pcadf = pd.read_csv(fname, header = None)
    df["0"] = pcadf.iloc[:,1]
    return df 



def read_all_colvars(fpattern, colname, colvars):
    res = []
    for colvar in colvars:
        fname = fpattern.replace("PLACEHOLDER", str(colvar))
        try: 
            df = read_algo_file(fpattern, colvar)
            df[colname] = colvar
            res.append(df)
            
        except FileNotFoundError:
            print("Can not find file", fname)
            pass 
    return pd.concat(res)

def boxlineplot(df, hue_colname, xaxisname, sortedxtickLabel, trial_colname = "trial", \
    value_colname = "distance", add_line = True, ax = None, **kwargs):

    g = df.groupby(hue_colname)
    ngroups = len(g)
    nlabels = len(sortedxtickLabel)

    boxwidthr = kwargs.get("barwidthr", 1)
    boxwidth = boxwidthr * 1/ngroups
    bp_colors = kwargs.get("boxcolors", dict(zip(g.groups, plt.get_cmap('Set1')(range(ngroups)))))

    algo_names = {"sriamp": "Bayes-OAMP", "riunamp": "single-iterate Bayes-OAMP", "ggamp": "Gaussian Bayes-AMP"}

    # for i, gkey in enumerate(g.groups):
    # for i, gkey in enumerate(["gamp", "ggamp", "riamp", "riunamp"]):
    for i, (gkey, group) in enumerate(g):
        # gdf = g.get_group(gkey).drop(hue_colname, axis = 1)
        gdf = group.drop(hue_colname, axis = 1)
        gdf = gdf.pivot(index = trial_colname, columns = xaxisname, values = value_colname).reindex(sortedxtickLabel, axis = 1)
        gdf.dropna(axis = 0, inplace = True)
        pos = getPos(nlabels, ngroups, i, boxwidth)
        box = ax.boxplot(gdf, positions = pos, sym='', widths = boxwidth)
        set_box_color(box, bp_colors[gkey])
        if add_line:
            # means = gdf.mean(axis = 0)
            ax.plot(np.arange(nlabels), gdf.mean(axis = 0), c = bp_colors[gkey], marker = ".")
            # counts = gdf.count()
            # write_counts(ax, counts, pos, means, bp_colors[i], 2*(i%2)-1)

    return ax

def newboxplot(fpattern, colname, colvars, figname):

    xticklabelvars = list(map(str, range(MAXITER)))
    xaxisname = "iteration"

    fig, (vax, uax) = plt.subplots(nrows = 2, figsize = (len(xticklabelvars)/2,5), constrained_layout=True)
    # fig, (vax, uax) = plt.subplots(ncols = 2, figsize = (len(xticklabelvars),3), constrained_layout=True)

    for uv, ax in zip(("v", "u"),(vax, uax)):
        uvfpattern = fpattern.replace("DIRECTION", uv)
        df = read_all_colvars(uvfpattern, colname, colvars)
        df = df.melt(id_vars = [colname, "trial"], \
            value_vars = [cn for cn in df.columns if cn!= colname], \
            var_name = "iteration", value_name = "distance")
        

        ax = boxlineplot(df, 
            hue_colname= "method",
            xaxisname = xaxisname,
            sortedxtickLabel = xticklabelvars,
            ax = ax,
            barwidthr = 0.5,
            boxcolors = boxcolors
            )
        
        markedlocs = list(range(0, len(xticklabelvars), 1))
        ax.set_xticks(markedlocs)
        
        ax.set_xticklabels([lbl for i, lbl in enumerate(xticklabelvars) if i in markedlocs])
        # ax.legend()
        
        if uv == "u":
            ax.set_xticklabels([int(lbl)+1 for i, lbl in enumerate(xticklabelvars) if i in markedlocs])
            ax.set_xlabel("Iterations") 

        if "mse" in fpattern:
            ax.set_ylabel("MSE")
        
        if "subspace" in fpattern:
            ax.set_ylabel("Subspace Distance")
        
        ax.set_title(figTitles[args.noisetype] + " " + uv.capitalize())

    fpath = os.path.split(fpattern.replace("PLACEHOLDER", ""))[0] + "_"
    fpath = fpath.replace("records", "figures")
    mkdir(fpath)
    fname = fpath + figname + ".png"
    print(fname)
    fig.savefig(fname, dpi = 400)


if __name__=="__main__":
    import argparse
    import numpy as np
    
    parser = argparse.ArgumentParser(prog = 'CompareAlgo')
    parser.add_argument("--noisetype", type=str)
    parser.add_argument("--local", action="store_true")
    parser.add_argument("--s1", nargs = "?", type=float, const = 2.0, default=2.0)
    parser.add_argument("--s2", nargs = "?", type=float, const = 1.5, default=1.5)
    
    args = parser.parse_args()

    signals = [args.s1, args.s2]

    
    fpattern = "../records/mse/PLACEHOLDER_n{}/{:.2f}_{:.2f}_DIRECTION.csv".format(args.noisetype, signals[0], signals[1]) 
    colvars = ["sriamp", "riunamp", "ggamp"]
    newboxplot(fpattern, "method", colvars, "comAlgo{:.2f}_{:.2f}".format(signals[0],signals[1]))
    quit()  


