'''
==========
Date Aug 26, 2021
State Evolution
==========
This module help organize state evolution for
for rect case
1. riamp
2. riunamp
3. gamp
'''


from abc import ABC, abstractmethod

import numpy as np 
from numpy.linalg import matrix_power as matprod

def get_Theta(j, Phi, Psi, Delta, Gamma):
    Theta = np.zeros(Phi.shape)
    PhiPsi = Phi.dot(Psi)
    # part1
    for i in range(j+1):
        prod = np.linalg.matrix_power(PhiPsi, i).dot(Delta)
        prod = prod.dot(np.linalg.matrix_power(np.transpose(PhiPsi), j-i))
        Theta += prod 
    # part 2
    PhiGammaPhit = Phi.dot(Gamma).dot(np.transpose(Phi))
    for i in range(j):
        prod = np.linalg.matrix_power(PhiPsi, i)
        prod = prod.dot(PhiGammaPhit)
        prod = prod.dot(matprod(np.transpose(PhiPsi), j - 1 -i))
        Theta += prod
    return Theta

def get_Xi(j, Phi, Psi, Delta, Gamma):
    Xi = np.zeros(Phi.shape)
    PsiPhi = Psi.dot(Phi)
    # part 1
    for i in range(j+1):
        prod = np.linalg.matrix_power(PsiPhi, i).dot(Gamma)
        prod = prod.dot(np.linalg.matrix_power(np.transpose(PsiPhi), j-i))
        Xi += prod     
    # part 2
    PsiDeltaPsi = Psi.dot(Delta).dot(np.transpose(Psi))
    for i in range(j):
        prod = np.linalg.matrix_power(PsiPhi, i)
        prod = prod.dot(PsiDeltaPsi)
        prod = prod.dot(matprod(np.transpose(PsiPhi), j - 1 -i))
        Xi += prod
    return Xi 

def get_Theta(j, Phi, Psi, Delta, Gamma):
    Theta = np.zeros(Phi.shape)
    PhiPsi = Phi.dot(Psi)
    # part1
    for i in range(j+1):
        prod = np.linalg.matrix_power(PhiPsi, i).dot(Delta)
        prod = prod.dot(np.linalg.matrix_power(np.transpose(PhiPsi), j-i))
        Theta += prod 
    # part 2
    PhiGammaPhit = Phi.dot(Gamma).dot(np.transpose(Phi))
    for i in range(j):
        prod = np.linalg.matrix_power(PhiPsi, i)
        prod = prod.dot(PhiGammaPhit)
        prod = prod.dot(matprod(np.transpose(PhiPsi), j - 1 -i))
        Theta += prod
    return Theta


def decompose_Delta(Delta, K):
    Delta0 = np.zeros(Delta.shape)
    Delta1 = np.zeros(Delta.shape)
    Delta2 = np.zeros(Delta.shape)
    Delta2[:K,:K] = Delta[:K,:K]
    Delta1[K:,:K] = Delta[K:,:K]
    Delta0[K:,K:] = Delta[K:,K:]
    return Delta0, Delta1, Delta2 

def get_A(t, Phi, Psi, kappas, zetas, K):
    '''
    A0 associates with all debiasing terms doesn't involve u_0
    A1 associates with all debiasing terms involving u_0
    '''
    assert Phi.shape == Psi.shape
    A0 = np.zeros(Phi.shape)
    A1 = np.zeros(Phi.shape)
    prod = Psi
    for j in range(t+1): #
        A0 += kappas[j+1] * prod
        A1 += prod * zetas[j+1]
        prod = prod.dot(Phi).dot(Psi)
    A = np.zeros(Phi.shape)
    A[:,K:] = A0[:,K:]
    A[:,:K] = A1[:,:K]
    return A  

def get_B(t, Phi, Psi, kappas, zetas, K):
    '''
    B0 associates with all debiasing terms doesn't involve v_0
    B1 associates with all debiasing terms involving v_0
    '''
    B0 = np.zeros(Phi.shape)
    B1 = np.zeros(Phi.shape)
    prod = Phi
    for j in range(t+2): #
        B0 += kappas[j+1] * prod
        B1 +=  prod * zetas[j+1]
        prod = prod.dot(Psi).dot(Phi)
    B = np.zeros(Phi.shape)
    B[:,K:] = B0[:,K:]
    B[:,:K] = B1[:,:K]
    return B

class _StateEvolution(ABC):

    def __init__(self, K: int, niters: int, aspect_ratio: float, thetas, uthetas, vthetas, kappas, zetas, dzetas):
        '''
        zetas, dzetas: ndarray: (J*K)
        '''
        self.K = K
        self.aspect_ratio = aspect_ratio
        self.niters = niters
        self.thetas = thetas
        self.uthetas = uthetas 
        self.vthetas = vthetas
        self.kappas = kappas
        self.zetas = np.tile(zetas, (1, niters+1))
        self.dzetas = np.tile(dzetas, (1, niters+1))
        self.Delta = np.zeros(((niters+1)*K, (niters+1)*K))
        self.Gamma = np.zeros(((niters+1)*K, (niters+1)*K))
        self.Phi = np.zeros(((niters+1)*K, (niters+1)*K))
        self.Phi[K:2*K,:K] = np.diag(1/uthetas)
        self.Psi = np.zeros(((niters+1)*K, (niters+1)*K))
        self.Psi[:K,:K] = np.diag(1/vthetas)
        self.shape = ((niters+1)*K, (niters+1)*K)


    @abstractmethod
    def getb(self,t):
        pass 

    @abstractmethod
    def geta(self,t):
        pass 

    @abstractmethod
    def getSigma(self, t):
        pass

    @abstractmethod
    def updatePhi(self, t, phiRow):
        pass

    @abstractmethod
    def updateDelta(self, t, Delta):
        pass

    @abstractmethod
    def updatePsi(self, t, psiRow):
        pass 

    @abstractmethod
    def updateGamma(self, t, Gamma):
        pass


class RISE(_StateEvolution):
    def __init__(self, K: int, niters: int, aspect_ratio: int, thetas, uthetas, vthetas, kappas, zetas, dzetas, *args, **kwargs):
        _StateEvolution.__init__(self, K, niters, aspect_ratio, thetas, uthetas, vthetas, kappas, zetas, dzetas)

    def updateGamma(self, t, Gamma):
        self.Gamma[:(t+1)*self.K, :(t+1)*self.K] = Gamma

    def updateDelta(self, t, Delta):
        self.Delta[:(t+1)*self.K, :(t+1)*self.K] = Delta

    def updatePhi(self, t, phiRow):
        self.Phi[(t*self.K):(t+1)*self.K, :(t*self.K)] = phiRow

    def updatePsi(self, t, psiRow):
        self.Psi[(t*self.K):(t+1)*self.K, self.K:(t+1)*self.K] = psiRow
    
    def geta(self, t):
        A = get_A(t, self.Phi, self.Psi, self.kappas, self.zetas, self.K)
        return A[(t*self.K):(t+1)*self.K] 
    
    def getb(self,t):
        B = get_B(t, self.Phi, self.Psi, self.kappas, self.zetas, self.K)
        return B[t*self.K:(t+1)*self.K]

    
    def getSigma(self, t):
        
        Delta0, Delta1, Delta2 = decompose_Delta(self.Delta, self.K)
        Gamma0, Gamma1, Gamma2 = decompose_Delta(self.Gamma, self.K)
        
        Sigma = np.zeros(self.shape)
        for j in range(2*t + 1):
            Deltaj =  Delta0*self.kappas[j+1] + Delta1*self.zetas[j+1] + (Delta1*self.zetas[j+1]).T + Delta2*self.dzetas[j+1] 
            Gammaj = Gamma0*self.kappas[j+1] + Gamma1*self.zetas[j+1] + (Gamma1*self.zetas[j+1]).T + Gamma2*self.dzetas[j+1] 
            Sigma += get_Xi(j, self.Phi, self.Psi, Deltaj, Gammaj)
        
        return Sigma[:(t+1)*self.K, :(t+1)*self.K]
    
    def getMut(self, t):
        return self.Gamma[t*self.K:(t+1)*self.K,t*self.K:(t+1)*self.K] * self.thetas / np.sqrt(self.aspect_ratio)


    def getOmega(self, t):
        
        Delta0, Delta1, Delta2 = decompose_Delta(self.Delta, self.K)
        Gamma0, Gamma1, Gamma2 = decompose_Delta(self.Gamma, self.K)
        
        Omega = np.zeros(self.shape)
        for j in range(2*t + 1):
            Deltaj =  Delta0*self.kappas[j+1] + Delta1*self.zetas[j+1] + (Delta1*self.zetas[j+1]).T + Delta2*self.dzetas[j+1] 
            Gammaj = Gamma0*self.kappas[j+1] + Gamma1*self.zetas[j+1] + (Gamma1*self.zetas[j+1]).T + Gamma2*self.dzetas[j+1] 
            Omega += get_Theta(j, self.Phi, self.Psi, Deltaj, Gammaj)
        
        return Omega[self.K:(t+1)*self.K, self.K:(t+1)*self.K]

    def getNut(self, t):
        return self.Delta[t*self.K:(t+1)*self.K,t*self.K:(t+1)*self.K] * self.thetas * np.sqrt(self.aspect_ratio)



class GSE(RISE):

    def __init__(self, K: int, niters: int, aspect_ratio: int, thetas, uthetas, vthetas, *args, **kwargs):
        
        kappas = None # np.array([0,0,1] + [0]*(niters*4-2))
        zetas = None # np.zeros((niters*4+1, K))
        # zetas[0] = 1/thetas**2; zetas[1] = 1/thetas; zetas[2] = 1
        dzetas = None # np.zeros((niters*4+1, K))
        # dzetas[0] = 3/thetas**2; dzetas[1] = 2/thetas; dzetas[2] = 1

        RISE.__init__(self, K, niters, aspect_ratio, thetas, uthetas, vthetas, kappas, zetas, dzetas)
    
    def getSigma(self, t):
        Sigma = self.Gamma
        return Sigma[:(t+1)*self.K, :(t+1)*self.K]

    def getOmega(self, t):
        Omega = self.Delta
        return Omega[self.K:(t+1)*self.K, self.K:(t+1)*self.K]
    
    def geta(self, t):
        A = self.Psi
        return A[(t*self.K):(t+1)*self.K] 
    
    def getb(self, t):
        B = self.Phi
        return B[t*self.K:(t+1)*self.K]


    
class SRISE(RISE):
    def __init__(self, K: int, niters: int, aspect_ratio: int, thetas, uthetas, vthetas,kappas, zetas, dzetas):
        self.reg = 1e-5
        RISE.__init__(self, K, niters, aspect_ratio, thetas, uthetas, vthetas, kappas, zetas, dzetas)
    
    def getSigma(self, t):
        Sigma = super().getSigma(t)
        # Regularization
        D,Q = np.linalg.eigh(Sigma)
        print("Sigma D at t={}".format(t))
        print(D)
        if D.min() < self.reg:
            raise RuntimeError("Hit stopping threshould")
        
        return Sigma 

    def getOmega(self, t):
        Omega = super().getOmega(t)
        # Regularization
        D,Q = np.linalg.eigh(Omega)
        print("Omega D at t={}".format(t))
        print(D)
        if D.min() < self.reg:
            raise RuntimeError("Hit stopping threshould")
        
        return Omega
    

class RRISE(RISE):
    def __init__(self, K: int, niters: int, aspect_ratio: int, thetas, uthetas, vthetas,kappas, zetas, dzetas):
        self.reg = 0
        RISE.__init__(self, K, niters, aspect_ratio, thetas, uthetas, vthetas, kappas, zetas, dzetas)
    
    def getSigma(self, t):
        Sigma = super().getSigma(t)
        # Regularization
        D,Q = np.linalg.eigh(Sigma)
        D = np.maximum(D,self.reg)
        Sigma = Q.dot(np.diag(D)).dot(np.transpose(Q))
        
        return Sigma 

    def getOmega(self, t):
        Omega = super().getOmega(t)
        # Regularization
        D,Q = np.linalg.eigh(Omega)
        D = np.maximum(D,self.reg)
        Omega = Q.dot(np.diag(D)).dot(np.transpose(Q))
        
        return Omega
    






    



        