'''
==========
Date: Jul 7, 2021
AMP algorithm with PCA initialization for symmetric matrices with higher rank
1. riamp
2. riunamp
3. gamp
==========
'''
from pca import SymPcaPack, get_gaussian_pcapack_from_pcapack

import numpy as np
np.seterr(all = "raise")

from se import *
from empbayes import NonparEB

from utils import get_avgfro_norm


def riamp(pcapack, denoiser = NonparEB(), niters = 5, SeModel = RISE, sePar = 0):
    
    Y = pcapack.Y
    thetas = pcapack.thetas

    n = Y.shape[0]
    K = len(thetas)

    # Init StateEvolution
    SE = SeModel(K, niters, pcapack.thetas, pcapack.kappas, pcapack.zetas, pcapack.dzetas, sePar)

    # Init Iterates
    f = F = np.reshape(pcapack.F, (-1,K)) # (f_0, ..., f_{T-1})
    U = np.zeros((n,(niters+1)*K)) #(u_0, ..., u_T)
    U[:,0:K] = F/thetas

    # Init States
    Mu = mut = np.diag(pcapack.innerprods)
    Sigma = sigmat = np.identity(K) - mut**2

    denoiser.fit(f, mut, sigmat, figname = str(0))

    for t in range(1, niters + 1):  

        print("Iter {}".format(t))

        try: 
            # Denoise
            u = denoiser.denoise(F[:, :t*K], Mu, Sigma)
            U[:,(t*K):(t*K+K)] = u 
        
        except (Exception, Warning, RuntimeWarning) as e:
            print(e)
            return U[:, :(t)*K], t-1
            
        try:
            if t == niters:
                return U, t

            gradu = denoiser.ddenoise(F[:,:t*K], Mu, Sigma)
            SE.updatePhi(t, gradu.mean(axis = 0))

            # Update State
            Delta = (U.T @ U/n)[:(t+1)*K, :(t+1)*K]
            # print("Delta")
            # print(Delta)
            SE.updateDelta(t, Delta)


            # Next Iteration
            b = SE.getb(t)
            f = Y.dot(u) - U.dot(b.T) 
            F = np.hstack((F,f))

            # Next State
            Sigma = SE.getSigma(t)
            # print("Sigma")
            # print(Sigma)
            sigmat = Sigma[-K:,-K:]
            # use empirical estimate might have auto correction effect
            mut = SE.getMut(t)
            # mut = np.transpose(u).dot(u)/n * thetas
            Mu = np.vstack((Mu, mut))

            denoiser.fit(f, mut, sigmat, figname = str(t))
        
        except (Exception, Warning, RuntimeWarning) as e:
            print(e)
            return U[:, :(t+1)*K], t
        

def rriamp(pcapack, denoiser = NonparEB(), niters = 5, SeModel = RRISE):
    return riamp(pcapack, denoiser, niters, SeModel=RRISE)

def sriamp(pcapack, denoiser = NonparEB(), niters = 5, SeModel = SRISE):
    return riamp(pcapack, denoiser, niters, SeModel=SRISE)
        

def eriamp(pcapack, ustar, denoiser = NonparEB(), niters = 5, reg = 1e-6, **kwargs):
    if "diagnosis" in kwargs:
        diagnosis = kwargs["diagnosis"]
    else:
        diagnosis = None

    Y = pcapack.Y
    thetas = pcapack.thetas

    n = Y.shape[0]
    K = len(thetas)

    # Init StateEvolution
    SE = RISE(K, niters, pcapack.thetas, pcapack.kappas, pcapack.zetas, pcapack.dzetas)

    # Init Iterates
    f = F = np.reshape(pcapack.F, (-1,K)) # (f_0, ..., f_{T-1})
    U = np.zeros((n,(niters+1)*K)) #(u_0, ..., u_T)
    U[:,0:K] = F/thetas

    # Init States
    Mu = mut = np.diag(pcapack.innerprods)
    Sigma = sigmat = np.identity(K) - mut**2

    denoiser.fit(f, mut, sigmat, figname = str(0))

    for t in range(1, niters + 1):
        print("At step", t)

        D,Q = np.linalg.eigh(Sigma)
        if D.min() < reg: 
            print("I stopped at Dmin reg")
            return U[:, :t*K], t-1

        try: 
            # Diagnosis 
            if diagnosis:
                mut = Mu[-K:,:]
                sampleMut = (F.T @ ustar / n) [-K:,:]
                diagnosis.record("mu", get_avgfro_norm(mut - sampleMut), t)
                sampleMu = (U.T @ ustar / n * thetas) [:K*t, :K*t]
                Mean = ustar @ sampleMu.T
                sampleSigma = np.cov(F - Mean, rowvar = False)
                diagnosis.record("cov", get_avgfro_norm(Sigma - sampleSigma), t)
                diagnosis.record("cond", np.linalg.cond(Sigma), t)
                diagnosis.record("invCov", get_avgfro_norm(np.linalg.inv(Sigma) - np.linalg.inv(sampleSigma)), t)
            
            # Denoise
            u = denoiser.denoise(F[:, :t*K], Mu, Sigma)
            U[:,(t*K):(t*K+K)] = u 
        
        except Exception as e:
            print("I stopped at denoise")
            print(e)
            return U[:, :t*K], t-1

        try:
            if diagnosis:
                diagnosis.record("dist", get_l2_dist(u, ustar), t)

            if t == niters:
                return U, t
        

            # Update State
            Delta = (U.T @ U/n)[:(t+1)*K, :(t+1)*K]
            SE.updateDelta(t, Delta)
            gradu = denoiser.ddenoise(F[:,:t*K], Mu, Sigma)
            SE.updatePhi(t, gradu.mean(axis = 0))

            # Next Iteration
            b = SE.getb(t)
            f = Y.dot(u) - U.dot(b.T) 
            F = np.hstack((F,f))

            # Next State
            Sigma = SE.getSigma(t)
            sigmat = Sigma[-K:,-K:]
            # use empirical estimate might have auto correction effect
            mut = SE.getMut(t)
            # mut = np.transpose(u).dot(u)/n * thetas
            Mu = np.vstack((Mu, mut))

            denoiser.fit(f, mut, sigmat, figname = str(t))
        
        except Exception as e:
            print("Error 2")
            print(e)
            return U[:, :(t+1)*K], t


def riunamp(pcapack, denoiser = NonparEB(), niters = 5, **kwargs):
    
    Y = pcapack.Y
    thetas = pcapack.thetas

    n = Y.shape[0]
    K = len(thetas)

    # Init StateEvolution
    SE = RISE(K, niters, pcapack.thetas, pcapack.kappas, pcapack.zetas, pcapack.dzetas)

    # Init Iterates
    f = F = np.reshape(pcapack.F, (-1,K)) # (f_0, ..., f_{T-1})
    U = np.zeros((n,(niters+1)*K)) #(u_0, ..., u_T)
    u = U[:,0:K] = F/thetas
    

    # Check init 
    mut = np.diag(pcapack.innerprods)
    sigmat = np.identity(K) - mut**2
    denoiser.fit(F, mut, sigmat, figname = str(0))

    for t in range(1, niters + 1):

        try: 
            # Denoise
            u = denoiser.denoise(f, mut, sigmat)
            U[:,(t*K):(t*K+K)] = u 
        
        except (Exception, Warning, RuntimeWarning) as e:
            print(e)
            return U[:, :(t)*K], t-1

        try:
            if t == niters:
                return U, t
            
            # Update State
            Delta = (U.T @ U/n)[:(t+1)*K, :(t+1)*K]
            SE.updateDelta(t, Delta)
            gradu = denoiser.ddenoise(f, mut, sigmat)
            phiRow = np.hstack((np.zeros(shape = (K, K*(t-1))), \
                gradu.mean(axis = 0) ))
            SE.updatePhi(t, phiRow)

            # Next Iteration
            b = SE.getb(t)
            f = Y.dot(u) - U.dot(b.T) 
            F = np.hstack((F,f))

            # Next State
            sigmat = SE.getSigma(t)[-K:,-K:]
            mut = SE.getMut(t)
            
            denoiser.fit(f, mut, sigmat, figname = str(t))
        
        except (Exception, Warning, RuntimeWarning) as e:
            print(e)
            return U[:, :(t)*K], t-1


def riunampD(pcapack, ustar, denoiser = NonparEB(), niters = 5, **kwargs):
    
    diagnosis = kwargs["diagnosis"]
    
    Y = pcapack.Y
    thetas = pcapack.thetas

    n = Y.shape[0]
    K = len(thetas)

    # Init StateEvolution
    SE = RISE(K, niters, pcapack.thetas, pcapack.kappas, pcapack.zetas, pcapack.dzetas)

    # Init Iterates
    f = F = np.reshape(pcapack.F, (-1,K)) # (f_0, ..., f_{T-1})
    U = np.zeros((n,(niters+1)*K)) #(u_0, ..., u_T)
    u = U[:,0:K] = F/thetas
    

    # Check init 
    mut = np.diag(pcapack.innerprods)
    sigmat = np.identity(K) - mut**2
    denoiser.fit(F, mut, sigmat, figname = str(0))

    for t in range(1, niters + 1):

        try: 
            # Diagnosis 
            sampleMut = (f.T @ ustar / n) [-K:,:]
            diagnosis.record("mu", get_avgfro_norm(mut - sampleMut), t)
            Mean = ustar @ sampleMut.T
            sampleSigmat = np.cov(f - Mean, rowvar = False)
            diagnosis.record("cov", get_avgfro_norm(sigmat - sampleSigmat), t)
            diagnosis.record("cond", np.linalg.cond(sigmat), t)
            diagnosis.record("invCov", get_avgfro_norm(np.linalg.inv(sigmat) - np.linalg.inv(sampleSigmat)), t)

            u = denoiser.denoise(f, mut, sigmat)
            U[:,(t*K):(t*K+K)] = u 

            diagnosis.record("dist", get_l2_dist(u, ustar), t)
            
            if t == niters:
                return U, t
            
            # Update State
            Delta = (U.T @ U/n)[:(t+1)*K, :(t+1)*K]
            SE.updateDelta(t, Delta)
            gradu = denoiser.ddenoise(f, mut, sigmat)
            phiRow = np.hstack((np.zeros(shape = (K, K*(t-1))), \
                gradu.mean(axis = 0) ))
            SE.updatePhi(t, phiRow)

            # Next Iteration
            b = SE.getb(t)
            f = Y.dot(u) - U.dot(b.T) 
            F = np.hstack((F,f))

            # Next State
            sigmat = SE.getSigma(t)[-K:,-K:]
            mut = SE.getMut(t)
            
            denoiser.fit(f, mut, sigmat, figname = str(t))
        
        except Exception:
            return U[:, :(t+1)*K], t
        
def gamp(pcapack, denoiser = NonparEB(), niters = 5, **kwargs):

    Y = pcapack.Y # NOTE: ensure Y is normalized
    thetas = pcapack.thetas

    n = Y.shape[0]
    K = len(thetas)

    # Init StateEvolution
    SE = GSE(K, niters, pcapack.thetas) 

    # Init Iterates
    f = F = np.reshape(pcapack.F, (-1,K)) # (f_0, ..., f_{T-1})
    U = np.zeros((n,(niters+1)*K)) #(u_0, ..., u_T)
    u = U[:,0:K] = F/thetas

    # Check init 
    mut = np.diag(pcapack.innerprods)
    sigmat = np.identity(K) - mut**2
    denoiser.fit(F, mut, sigmat, figname = str(0))

    for t in range(1, niters + 1):

        print("Gamp iter {}".format(t))
        
        try:
            u = denoiser.denoise(f, mut, sigmat)
            U[:,(t*K):(t*K+K)] = u 

        except (Exception, Warning, RuntimeWarning) as e:
            print(e)
            return U[:, :(t)*K], t-1
            
        try:

            if t == niters:
                return U, t
            
            # Update State
            Delta = (U.T @ U/n)[:(t+1)*K, :(t+1)*K]
            SE.updateDelta(t, Delta)
            gradu = denoiser.ddenoise(f, mut, sigmat)
            phiRow = np.hstack((np.zeros(shape = (K, K*(t-1))), \
                gradu.mean(axis = 0) ))
            SE.updatePhi(t, phiRow)

            # Next Iteration
            b = SE.getb(t)
            f = Y.dot(u) - U.dot(b.T) 
            F = np.hstack((F, f))

            # Next State
            sigmat = SE.getSigma(t)[-K:,-K:]
            mut = SE.getMut(t)
            
            denoiser.fit(f, mut, sigmat, figname = str(t))
        
        except (Exception, Warning, RuntimeWarning) as e:
            print(e)
            return U[:, :(t+1)*K], t

def ggamp(opcapack, denoiser = NonparEB(), niters = 5, **kwargs):
    
    pcapack = get_gaussian_pcapack_from_pcapack(opcapack)

    return gamp(pcapack, denoiser=denoiser, niters = niters, **kwargs)



def gampD(pcapack, ustar, denoiser = NonparEB(), niters = 5, **kwargs):
    
    diagnosis = kwargs["diagnosis"]

    Y = pcapack.Y # NOTE: ensure Y is normalized
    thetas = pcapack.thetas

    n = Y.shape[0]
    K = len(thetas)

    # Init StateEvolution
    SE = GSE(K, niters, pcapack.thetas)

    # Init Iterates
    f = F = np.reshape(pcapack.F, (-1,K)) # (f_0, ..., f_{T-1})
    U = np.zeros((n,(niters+1)*K)) #(u_0, ..., u_T)
    u = U[:,0:K] = F/thetas

    # Check init 
    mut = np.diag(pcapack.innerprods)
    sigmat = np.identity(K) - mut**2
    denoiser.fit(F, mut, sigmat, figname = str(0))

    for t in range(1, niters + 1):
        
        try:
            # Diagnosis 
            sampleMut = (f.T @ ustar / n) [-K:,:]
            diagnosis.record("mu", get_avgfro_norm(mut - sampleMut), t)
            Mean = ustar @ sampleMut.T
            sampleSigmat = np.cov(f - Mean, rowvar = False)
            diagnosis.record("cov", get_avgfro_norm(sigmat - sampleSigmat), t)
            diagnosis.record("cond", np.linalg.cond(sigmat), t)
            diagnosis.record("invCov", get_avgfro_norm(np.linalg.inv(sigmat) - np.linalg.inv(sampleSigmat)), t)


            u = denoiser.denoise(f, mut, sigmat)
            U[:,(t*K):(t*K+K)] = u 

            diagnosis.record("dist", get_l2_dist(u, ustar), t)
            
            if t == niters:
                return U, t
            
            # Update State
            Delta = (U.T @ U/n)[:(t+1)*K, :(t+1)*K]
            SE.updateDelta(t, Delta)
            gradu = denoiser.ddenoise(f, mut, sigmat)
            phiRow = np.hstack((np.zeros(shape = (K, K*(t-1))), \
                gradu.mean(axis = 0) ))
            SE.updatePhi(t, phiRow)

            # Next Iteration
            b = SE.getb(t)
            f = Y.dot(u) - U.dot(b.T) 
            F = np.hstack((F, f))

            # Next State
            sigmat = SE.getSigma(t)[-K:,-K:]
            mut = SE.getMut(t)
            
            denoiser.fit(f, mut, sigmat, figname = str(t))
        except Exception:
            return U[:, :(t+1)*K], t

AMPALGO = {
    "riamp": riamp,
    "sriamp": sriamp, # this is with the new get W
    "riunamp": riunamp,
    "gamp": gamp,
    "rriamp": rriamp,
    "eriamp": eriamp,
    "ggamp": ggamp
}
