'''
==========
Date: Jul 8
Noise
Note we need to normalize all W such that kappa1 = 1
==========
This module provide 
1) noise generator
2) pca analysis on 
'''

from abc import ABC, abstractmethod

import numpy as np 
from numpy.polynomial import polynomial
import math 
from scipy.optimize import fsolve
from scipy.special import bernoulli

def _get_moments_from_empirical_measure(mu, K):
    m = []
    for j in range(K + 1):
        m.append(np.mean(mu**j))
    return np.array(m)

def _cauchy_transform(z, mu):
    return np.mean(1/(z-mu))

def _vec_cauchy_transform(z, mu):
    inner = lambda x: _cauchy_transform(x, mu)
    return np.vectorize(inner)(z)

def _inv_cauchy_transform(zs, mu, zinits):
    res = []
    for z, zinit in zip(zs, zinits):
        func = lambda x: _cauchy_transform(x, mu) - z
        sol = fsolve(func, zinit)
        res.append(sol)
    return np.array(res)

def _d_cauchy_transform(zs, mu):
    res = []
    for z in zs:
        res.append(-np.mean(1/np.square(z-mu)))
    return np.array(res)

def _phase_transition_threshold(mu, tol = 0.01):
    # upper threshold 
    supp_max_plus = mu.max() + tol
    rcp = 1/_cauchy_transform(supp_max_plus, mu)
    # lower threshold
    supp_min_minus = mu.min() - tol
    lcp = 1/_cauchy_transform(supp_min_minus,mu)

    return (lcp, rcp)

def _get_cumulants_from_moments(m,K):
    '''
    # TODO, check length 
    Return kappas[0] to be kappa_0
    Return kappas0, ... ,kappasK
    '''
    S = polynomial.polymul([0,1],m)
    kappa = [0,m[1]]
    for k in range(2,K+1):
        P = [0]
        try: 
            for j in range(1,k):
                P = polynomial.polyadd(P,kappa[j]*polynomial.polypow(S,j))
                # print("k = {}, j = {}, P =".format(k,j), P)
            if len(P) < k:
                kappa.append(m[k]-0)
            else:
                kappa.append(m[k]-P[k])
        except FloatingPointError:
            kappa.append(m[k] - 0)
    return kappa

class NoiseFactory:
    """ The factory class for all defined noises"""

    registry = {}
    """ Internal registry for available noises """

    @classmethod
    def register(cls, name: str):
        """ Class method to register Noise class to the internal registry.
        Args:
            name (str): The name of the noise.
        Returns:
            The specific noise class itself.
        """

        def inner_wrapper(wrapped_class):
            if name in cls.registry:
                print('%s noise already exists. Will replace it' % name)
            cls.registry[name] = wrapped_class
            return wrapped_class

        return inner_wrapper


class _Noise(ABC):

    __phaseTransition = None
    
    def __init__(self, n, seed = 1997):
        self.rng = np.random.default_rng(seed)
        self.W = self.generateW(n)
    
    @property
    def phaseTransition(self):
        if self.__phaseTransition is None:
            self.__phaseTransition = self.getPhaseTransition()
        return self.__phaseTransition

    @abstractmethod
    def getPhaseTransition(self):    
        pass
    
    @abstractmethod
    def getSingval(self, thetas):
        pass 
    
    @abstractmethod
    def getInnerprod(self, thetas):
        pass 
    
    @abstractmethod
    def generateW(self, n):
        pass     
    
    @abstractmethod
    def getCumulants(self, J):
        # Return k_1, ..., k_J
        pass         





@NoiseFactory.register("goe")
class GOE(_Noise):

    name = "goe"

    def __init__(self, n, seed = 1997):
        _Noise.__init__(self, n, seed)

    @staticmethod
    def getPhaseTransition():
        return 1
    
    @staticmethod
    def getSingval(thetas):
        return thetas + 1/thetas
    
    @staticmethod
    def getInnerprod(thetas):
        return np.sqrt(1 - 1/thetas**2)
    
    def generateW(self, n):
        W = self.rng.normal(size=(n,n))
        W = (W+np.transpose(W))/np.sqrt(2*n)
        return W 
    
    @staticmethod
    def getCumulants(J):
        return np.array([0,0,1] + [0]*(J-2))

@NoiseFactory.register("uniform")
class Uniform(_Noise):

    name = "uniform"
    # [-\sqrt{3}, \sqrt{3}]

    def __init__(self, n, seed = 1997):
        _Noise.__init__(self, n, seed)

    @staticmethod
    def _G(z):
        return np.log(np.abs(z + np.sqrt(3))/np.abs(z - np.sqrt(3)))/2/np.sqrt(3)
    
    @staticmethod
    def _invG(y):
        yprime = np.exp(2*np.sqrt(3)* y)
        z = (yprime + 1) * np.sqrt(3)/(yprime - 1)
        return z 
    
    @staticmethod
    def _dG(z):
        assert z.min()**2 > 3
        return -1/(z**2 - 3)

    @staticmethod
    def getPhaseTransition():
        print("Seems like 0")
        return 0
    
    @staticmethod
    def getSingval(thetas):
        return Uniform._invG(1/thetas)
    
    @staticmethod
    def getInnerprod(thetas):
        invG = Uniform._invG(1/thetas)
        dG = Uniform._dG(invG)
        return 1/thetas/np.sqrt(-dG)

    def generateW(self, n):
        mu = self.rng.uniform(-1,1,size = n)
        mu = mu/np.sqrt((mu**2).mean())
        W = self.rng.normal(size=(n,n))
        O = np.linalg.qr(W)[0]
        D = np.diag(mu)
        W = np.dot(np.dot(O,D),np.transpose(O))
        return W 
    
    @staticmethod
    def oldGetCumulants(J):
        moments = np.tile([1,0], reps=int(J/2)+1)
        scaler = [np.sqrt(3)**k/(k+1) for k in range(len(moments))]
        moments = moments * np.array(scaler)
        # print("moments")
        # print(moments)
        return _get_cumulants_from_moments(moments, J)
    
    @staticmethod
    def getCumulants(J):
        bn = bernoulli(J)
        bn = np.asarray([bn[j] if j%2 == 0 else 0 for j in range(J+1)])
        bn[0] = 0
        fac = np.asarray([math.factorial(j) for j in range(J+1)])
        scaler = np.asarray([math.pow((2*math.sqrt(3)), j) for j in range(J + 1)])
        return bn * scaler/ fac 

    


    



    


@NoiseFactory.register("rad")
class Rademacher(_Noise):

    name = "rad"

    def __init__(self, n, seed = 1997):
        _Noise.__init__(self, n, seed)

    @staticmethod
    def getPhaseTransition():
        return 0
    
    @staticmethod
    def _invG(y): 
        return (1 + np.sqrt(4*y**2 +1))/2/y
    
    @staticmethod
    def _Gprime(z):
        return - 1/(z - 1/z)**2*(1 + 1/z**2)
    
    @staticmethod
    def _Rprime(z):
        return 1/Rademacher._Gprime(Rademacher._invG(z)) + 1/z**2
    
    @classmethod
    def getSingval(cls, thetas):
        return Rademacher._invG(1/thetas)
    
    @staticmethod
    def getInnerprod(thetas):
        return np.sqrt(1 - Rademacher._Rprime(1/thetas)/thetas**2)

    def generateW(self, n):
        W = self.rng.normal(size=(n,n))
        O = np.linalg.qr(W)[0]
        D = np.diag(self.rng.binomial(1,0.5,size=n)*2-1)
        W = np.dot(np.dot(O,D),np.transpose(O))
        return W 

    @staticmethod
    def getCumulants(J):
        moments = np.tile([1,0], reps=int(J/2)+1)
        return _get_cumulants_from_moments(moments, J)




class _EmpiricalNoise(_Noise, ABC):
    
    def __init__(self, n, seed = 1997):
        self.rng = np.random.default_rng(seed)
        self.mu = self.getMu(n)
        _Noise.__init__(self, n, seed)
        self.singval = None 

    @abstractmethod
    def getMu(self, n):
        pass

    def generateW(self, n):
        mu = np.repeat(self.mu, int((n-1)/len(self.mu)) + 1)[:n]
        W = self.rng.normal(size=(n,n))
        O = np.linalg.qr(W)[0]
        D = np.diag(mu)
        W = np.dot(np.dot(O,D),np.transpose(O))
        return W 
    
    def getCumulants(self, J):
        moments = _get_moments_from_empirical_measure(self.mu, J)
        # print("moments")
        # print(moments)
        cumulants = _get_cumulants_from_moments(moments, J)
        print("cumulants")
        print(cumulants)
        return _get_cumulants_from_moments(moments, J)
    
    def getPhaseTransition(self):
        return _phase_transition_threshold(self.mu)
    
    def getSingval(self, thetas, empLambdas = None):
        
        if self.singval is not None:
            return self.singval
        # else 
        if empLambdas is None:
            empLambdas = thetas + 1/thetas
        
        self.singval =  _inv_cauchy_transform(1/thetas, self.mu, empLambdas).flatten()

        return self.singval

    
    def getInnerprod(self, thetas):
        # all values evaluated at 1/thetas
        invG = self.getSingval(thetas)
        print("invG", invG)
        dG = _d_cauchy_transform(invG, self.mu)
        print("dG", dG)
        # dR = 1/dG + thetas**2 
        return 1/np.sqrt(-dG)/thetas


@NoiseFactory.register("empbimodal")
class Bimodal(_EmpiricalNoise):

    name = "empbimodal"
    
    def __init__(self, n, **kwargs):
        _EmpiricalNoise.__init__(self, n, kwargs.get("seed", 1997))
    
    def getMu(self, n):
        mu = self.rng.binomial(1,0.5,size=n)*2-1
        mu = np.clip(mu + np.random.normal(size = n)/5,-2,2)
        return mu/np.sqrt((mu**2).mean())

@NoiseFactory.register("normal")
class Normal(_EmpiricalNoise):

    name = "normal"
    
    def __init__(self, n, **kwargs):
        _EmpiricalNoise.__init__(self, n, kwargs.get("seed", 1997))
    
    def getMu(self, n):
        mu = np.diag(self.rng.binomial(1,0.5,size=n)*2-1)
        mu = np.clip(self.rng.normal(size = n),-2,2)
        return mu/np.sqrt((mu**2).mean())

@NoiseFactory.register("empuniform")
class EmpUniform(_EmpiricalNoise):

    name = "empuniform"
    
    def __init__(self, n, **kwargs):
        _EmpiricalNoise.__init__(self, n, kwargs.get("seed", 1997))
    
    def getMu(self, n):
        mu = self.rng.uniform(-1,1,size = n)
        return mu/np.sqrt((mu**2).mean())

@NoiseFactory.register("centeredBeta")
class Normal(_EmpiricalNoise):

    name = "centeredBeta"
    
    def __init__(self, n, **kwargs):
        _EmpiricalNoise.__init__(self, n, kwargs.get("seed", 1997))
    
    def getMu(self, n):
        mu = self.rng.beta(3,1,size=n) # 13 very good  #  Beta 3,3
        mu -= np.mean(mu)
        return mu/np.sqrt((mu**2).mean())

@NoiseFactory.register("empgoe")
class empGOE(_EmpiricalNoise):

    name = "empgoe"
    
    def __init__(self, n, **kwargs):
        _EmpiricalNoise.__init__(self, n, kwargs.get("seed", 1997))
    
    def getMu(self, n):
        W = self.rng.normal(size=(n,n))
        W = (W+np.transpose(W))/np.sqrt(2*n)
        mu = np.linalg.eigvals(W)
        return mu/np.sqrt((mu**2).mean())

@NoiseFactory.register("emprad")
class empRAD(_EmpiricalNoise):

    name = "emprad"
    
    def __init__(self, n, **kwargs):
        _EmpiricalNoise.__init__(self, n, kwargs.get("seed", 1997))
    
    def getMu(self, n):
        mu = self.rng.binomial(1, 0.5, size = n) * 2 -1
        return mu  

@NoiseFactory.register("emptriangle")
class Triangle(_EmpiricalNoise):

    name = "emptriangle"

    # def getPhaseTransition():
    #     return 2

    def __init__(self, n, **kwargs):
        _EmpiricalNoise.__init__(self, n, kwargs.get("seed", 1997))
    
    def getMu(self, n):
        mu = self.rng.uniform(-1,1,size = n)
        mu = (1 - np.sqrt(1-mu))*(mu > 0) + (np.sqrt(1 + mu) - 1)*(mu<0)
        return mu  

class NoiseChef():
    
    def __init__(self, mu):
        self.thetas = None
        self.singvals = None 
        self.dCauchy = None
        self.mu = mu

    def estimateThetas(self, singvals, tol = 0):
        self.singvals = singvals
        supp_max = self.mu.max()       
        if singvals.min() <= (supp_max+tol):
            print("singval = {}; mu_supp = {}".format(np.array(singvals).min(), supp_max+tol))
            raise ValueError("signal must seperate from singval supp of noise matrix")
        self.thetas = 1/_vec_cauchy_transform(self.singvals, self.mu)        
        self.dCaucy = _d_cauchy_transform(self.singvals, self.mu)        
        return self.thetas 

    def checkInfo(self, singvals):
        if self.singvals is None:
            if singvals is not None:
                _ = self.estimateTheta(singvals)
            else: 
                raise RuntimeError("Without singvals we can not estimate.")
        if (self.singvals is not None) and ( np.array_equal(self.singvals, singvals)):
            raise Warning("Singvals are not consisitent. Update with the latest")
    
    # TODO Phasetranstion tide up
    def estimatePhaseTransition(self):
        return _phase_transition_threshold(self.mu)
    

    def estimateInnerprods(self, singvals = None):
        self.checkInfo(singvals)
        return np.sqrt(-1/self.dCaucy)/self.thetas
    
    def getCumulants(self, J):
        moments = _get_moments_from_empirical_measure(self.mu, J)
        # cumulants = _get_cumulants_from_moments(moments, J)
        return _get_cumulants_from_moments(moments, J)
    

