from collections import namedtuple

from noise import NoiseChef
import numpy as np

SymPcaPack = namedtuple("SymPcaPack",["Y", "F", \
    "thetas", "singvals", "innerprods", \
        "kappas", "zetas", "dzetas"])


def get_zetas_from_kappas(zeta0, theta, kappas):
    '''
    Return zetas[i] to be zeta_i
    '''
    zetas = [zeta0]
    prev = zeta0
    for i in range(len(kappas)):
        prev = (prev - kappas[i])*theta
        zetas.append(prev)
    return np.array(zetas)

def get_dzetas_from_zetas(dzeta0, theta, zetas):
    '''
    Return dzetas[i] to be dzetas_i
    '''
    dzetas = [dzeta0]
    prev = dzeta0
    for i in range(len(zetas)):
        prev = (prev - zetas[i])*theta
        dzetas.append(prev)
    return np.array(dzetas)

def get_zeta0(lmda, theta, kappa0):
    return lmda/theta - 1 + kappa0

def get_dzeta0(mu_pca, zeta0, kappa0):
    return 1 - mu_pca**2 - kappa0 + 2*zeta0

def get_emppcapack(Y, rank, niters):
    n = Y.shape[0]
    J = 2*niters + 2
    
    w, v = np.linalg.eigh(Y)
    # High to low 
    singvals = np.flip(w[-rank:])
    fsvd = np.fliplr(v[:,-rank:] * np.sqrt(n))
    
    noiseChef = NoiseChef(w[:n-rank])
    print("Phase Transition Threshold", noiseChef.estimatePhaseTransition())

    thetas = noiseChef.estimateThetas(singvals)
    print("Estimated Thetas:", thetas)
    innerprods = noiseChef.estimateInnerprods()
    print("Estimated Innerprods", innerprods)
    
    kappas = noiseChef.getCumulants(J)

    zeta0 = get_zeta0(singvals, thetas, kappas[0]) # K ndarray
    zetas = get_zetas_from_kappas(zeta0, thetas, kappas) # J*K ndarray
    dzeta0 = get_dzeta0(innerprods, zetas[0], kappas[0])
    dzetas = get_dzetas_from_zetas(dzeta0, thetas, zetas)



    pcapack = SymPcaPack(Y = Y, F = fsvd, \
        thetas = thetas, singvals = singvals, innerprods = innerprods, \
            kappas = kappas, zetas = zetas, dzetas = dzetas)
    return pcapack



def get_pcapack(Y, thetas, noise, niters):
    n,_ = Y.shape
    rank = len(thetas)
    J = 4*niters 

    thetas = np.array(thetas)
    # get top K eigenvalues and eigenvectors
    w, v = np.linalg.eigh(Y)
    # singvals = w[-rank:][::-1] # descending
    fsvd = np.fliplr(v[:,-rank:] * np.sqrt(n))

    singvals, innerprods = noise.getSingval(thetas), noise.getInnerprod(thetas)
    print("Analytical innerprods", innerprods)

    kappas = noise.getCumulants(J)

    zeta0 = get_zeta0(singvals, thetas, kappas[0]) # K ndarray
    zetas = get_zetas_from_kappas(zeta0, thetas, kappas) # J*K ndarray
    dzeta0 = get_dzeta0(innerprods, zetas[0], kappas[0])
    dzetas = get_dzetas_from_zetas(dzeta0, thetas, zetas)

    pcapack = SymPcaPack(Y = Y, F = fsvd, \
        thetas = thetas, singvals = singvals, innerprods = innerprods, \
            kappas = kappas, zetas = zetas, dzetas = dzetas)

    return pcapack

def get_gaussian_pcapack_from_pcapack(opcapack: SymPcaPack):

    singvals = opcapack.singvals
    print("Singvals are", singvals)
    # print("phase transition are", 2)

    thetas = (singvals + np.sqrt(singvals**2 - 4)) / 2 
    print("GAMP estimated thetas", thetas)
    ualigns = np.sqrt(1 - 1/thetas**2)
    

    pcapack = SymPcaPack(Y = opcapack.Y, F = opcapack.F, \
        thetas = thetas, singvals = opcapack.singvals, innerprods = ualigns, \
            kappas = 0, zetas = 0, dzetas = 0)

    return pcapack