import os 

import pandas as pd 
import numpy as np
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt

from utils import mkdir 

NITER = 5

from vis import getPos, set_box_color

figTitles = {
    "empgoe": "Semi-circle",
    "centeredBeta": "Centered Beta",
    "emprad": "Rademacher",
    "empuniform": "Uniform"
}

boxcolors = {
    "sriamp": "red",
    "riunamp": "#069AF3",
    "pca": "gray",
    "ggamp": "green"
}


def read_amp_result(fname): # MODIFIED
    df = pd.read_csv(fname, header = None)
    df = df.iloc[:, [NITER]].tail(50)
    df.columns = ["distance"]
    df["trial"] = np.arange(len(df))
    return df

def read_pca_result(fname):
    df = pd.read_csv(fname, header = None)
    df = df.iloc[:, [-1]].tail(50)
    df.columns = ["distance"]
    df["trial"] = np.arange(len(df))
    return df


def read_all_huevars_xnames(fpattern, huename, huevars, xaxisname, xaxisvars):
    res = []
    for huevar in huevars:
        # if huevar == "pca":
        #     continue
        fxname = fpattern.replace("PLACEHOLDER", str(huevar))
        for xaxisvar in xaxisvars:
            fname = fxname.replace("XAXISNAME", str(xaxisvar))
            try: 
                print(fname)
                if huevar == "pca":
                    df = read_pca_result(fname) 
                else: 
                    df = read_amp_result(fname)
                df[huename] = huevar
                df[xaxisname] = xaxisvar
                res.append(df)
                
                # if huevar == "sriamp":
                #     print(fname)
                #     df = read_pca_result(fname)
                #     df[huename] = "pca"
                #     df[xaxisname] = xaxisvar
                #     res.append(df)
                    
            except FileNotFoundError:
                print("Can not find file", fname)
                pass 
            except Exception as err:
                print("Err reading files\n{}".format(err))
    return pd.concat(res)


def boxlineplot(df, hue_colname, xaxisname, sortedxtickLabel, trial_colname = "trial", \
    value_colname = "distance", add_line = True, ax = None, **kwargs):

    g = df.groupby(hue_colname)
    ngroups = len(g)
    nlabels = len(sortedxtickLabel)

    boxwidthr = kwargs.get("barwidthr", 1)
    boxwidth = boxwidthr * 1/ngroups
    bp_colors = kwargs.get("boxcolors", dict(zip(g.groups, plt.get_cmap('Set1')(range(ngroups)))))

    algo_names = {"sriamp": "Bayes-OAMP", "riunamp": "single-iterate Bayes-OAMP", "ggamp": "Gaussian Bayes-AMP", "rriamp": "RRIAMP", "pca": "best linear estimator"}
    # algo_names = dict(zip(g.groups, g.groups))

    for i, (gkey, group) in enumerate(g):
        # gdf = g.get_group(gkey).drop(hue_colname, axis = 1)
        gdf = group.drop(hue_colname, axis = 1)
        gdf = gdf.pivot(index = trial_colname, columns = xaxisname, values = value_colname).reindex(sortedxtickLabel, axis = 1)
        # gdf.dropna(axis = 0, inplace = True)
        pos = getPos(nlabels, ngroups, i, boxwidth)
        box = ax.boxplot(gdf, positions = pos, sym='', widths = boxwidth)
        set_box_color(box, bp_colors[gkey])
        if add_line:
            ax.plot(np.arange(nlabels), gdf.median(axis = 0), c = bp_colors[gkey], marker = "")
            # counts = gdf.count()
            # write_counts(ax, counts, pos, means, bp_colors[i], 2*(i%2)-1)
        ax.plot([], c=bp_colors[gkey], label="{}".format(algo_names[gkey]))
    
    return ax


def newboxplot(fpattern, huename, huevars, figname):

    xvars = xaxisvars = ["{:.2f}".format(i) for i in np.arange(0.1,2.1,0.1)]
    xname = xaxisname = "theta"

    df = read_all_huevars_xnames(fpattern, huename, huevars, xaxisname, xaxisvars)
    
    fig, ax = plt.subplots(figsize = (6,3), constrained_layout=True)
    ax = boxlineplot(df, 
        hue_colname= "method",
        xaxisname= xname,
        sortedxtickLabel = xvars,
        ax = ax,
        barwidthr = 0.5,
        boxcolors = boxcolors
        )
    markedlocs = list(range(1, len(xvars), 2))
    # markedlocs = list(range(0, len(xvars)))
    ax.set_xticks(markedlocs)
    ax.set_xticklabels([lbl[:-1] for i, lbl in enumerate(xvars) if i in markedlocs])
    
    if "mse" in fpattern:
        ax.set_ylabel("MSE")
    else:
        ax.set_ylabel("Subspace Distance")
    ax.set_xlabel("Signal Strength")

    ax.legend()
    
    # ax.grid(True, linestyle='--', linewidth = 1)
    ax.set_title(figTitles[args.noisetype])
    fpath = os.path.split(fpattern.replace("PLACEHOLDER", ""))[0] + "_"
    fpath = fpath.replace("records", "figures")
    mkdir(fpath)
    fname = fpath + figname + ".png"
    print(fname) 
    
    fig.savefig(fname, dpi = 400)

    # labels:
    fpath = fpattern.rsplit('/', 2)[0]
    fpath = fpath.replace("records", "figures")
    fname = fpath + "/theta_algo_label" + ".png"
    if os.path.exists(fname):
        return 
    label_params = ax.get_legend_handles_labels() 
    figl, axl = plt.subplots(figsize = (1,3))
    axl.axis(False)
    axl.legend(*label_params, loc="center", ncol = len(huevars), bbox_to_anchor=(0.5, 0.5), prop={"size":10}, frameon=False)
    figl.savefig(fname, dpi = 400)
    print("label pos " + fname)

if __name__=="__main__":
    import argparse
    import numpy as np
    
    parser = argparse.ArgumentParser(prog = 'CompareAlgo')
    parser.add_argument("--noisetype", type=str)
    parser.add_argument("--local", action="store_true")
    parser.add_argument("--s1", nargs = "?", type=float, const = 2.0, default=2.0)
    parser.add_argument("--s2", nargs = "?", type=float, const = 1.6, default=1.6)
    
    args = parser.parse_args()

    signals = [args.s1, args.s2]

    if args.local == True:
        fpattern = "/Users/tracey/Desktop/0OAMP/hpc/records/PLACEHOLDER_n{}/{:.2f}_{:.2f}.csv".format(args.noisetype, signals[0], signals[1]) 
        huevars = ["riamp", "riunamp", "ggamp"]
        newboxplot(fpattern, "method", huevars, "comSignal")
        quit()  


    fpattern = "../records/subspaceDist/rank1/PLACEHOLDER_n{}/XAXISNAME.csv".format(args.noisetype) 
    huevars = ["pca", "ggamp", "riunamp", "sriamp"]
    newboxplot(fpattern, "method", huevars, "comSignaliter{}".format(NITER))







