'''
==========
Date: Jul 7, 2021
State Evolution
==========
This module help organize state evolution for
for symmetic case
1. riamp
2. riunamp
3. gamp
'''

from abc import ABC, abstractmethod

import numpy as np 



class _StateEvolution(ABC):

    def __init__(self, K, niters, thetas, kappas, zetas, dzetas):
        '''
        zetas, dzetas: ndarray: (J*K)
        '''
        self.K = K
        self.niters = niters
        self.thetas = thetas 
        self.kappas = kappas
        self.zetas = np.tile(zetas, (1, niters+1))
        self.dzetas = np.tile(dzetas, (1, niters+1))
        self.Delta = np.zeros(((niters+1)*K, (niters+1)*K))
        self.Phi = np.zeros(((niters+1)*K, (niters+1)*K))
        self.shape = ((niters+1)*K, (niters+1)*K)
        # Set Value
        self.Delta[:K,:K] = np.diag(1/self.thetas**2)




    @abstractmethod
    def _getB(self, t):
        pass

    @abstractmethod
    def getb(self,t):
        pass 

    @abstractmethod
    def getSigma(self, t):
        pass

    @abstractmethod
    def updatePhi(self, t, phiRow):
        pass

    @abstractmethod
    def updateDelta(self, t, Delta):
        pass


class RISE(_StateEvolution):
    
    def __init__(self, K, niters, thetas, kappas, zetas, dzetas, *args, **kwargs):
        _StateEvolution.__init__(self, K, niters, thetas, kappas, zetas, dzetas)
    
    @staticmethod
    def _decompose_Delta(Delta, K):
        Delta0 = np.zeros(Delta.shape)
        Delta1 = np.zeros(Delta.shape)
        Delta2 = np.zeros(Delta.shape)
        Delta2[:K,:K] = Delta[:K,:K]
        Delta1[K:,:K] = Delta[K:,:K]
        Delta0[K:,K:] = Delta[K:,K:]
        return Delta0, Delta1, Delta2 
    
    @staticmethod
    def _getTheta(j, Phi, Delta):
        Theta = np.zeros(Delta.shape)
        for i in range(j+1):
            prod = np.linalg.matrix_power(Phi,i).dot(Delta)
            prod = prod.dot(np.linalg.matrix_power(np.transpose(Phi),j-i))
            Theta += prod
        return Theta

    def _getB(self, t):
        '''
        B0 associates with all debiasing terms doesn't involve u_0
        B1 assciates with all debiasing terms invoving u_0
        '''
        B0 = np.zeros(self.shape)
        B1 = np.zeros(self.shape)
        B = np.zeros(self.shape)
        Phij = np.identity(self.shape[0]) # Phi to the power of j
        for j in range(t+1):
            B0 += Phij * self.kappas[j+1]
            B1 += Phij * self.zetas[j+1]
            Phij = Phij @ self.Phi
        B[:,self.K:] = B0[:,self.K:]
        B[:,:self.K] = B1[:,:self.K]
        return B
    
    def getb(self, t):
        B = self._getB(t)

        # Monitor 
        # print("B")
        # print(B[(t*self.K):(t+1)*self.K,:])
        # print("Phi")
        # print(self.Phi[(t*self.K):(t+1)*self.K,:])
        
        return B[(t*self.K):(t+1)*self.K,:] 

    def getMut(self, t):
        return self.Delta[t*self.K:(t+1)*self.K, t*self.K:(t+1)*self.K] * self.thetas


    def getSigma(self, t):
        '''
        0 indicates the term doen's involve u_0
        1 indicates the term involving 1 u_0
        2 indicates the term involving 2 u_0
        we get the Sigma for f0, ... , fT
        '''
        # TODO check length for kappas, zetas and dzetas.
        # Monitor

        Sigma = np.zeros(self.shape)

        Delta0, Delta1, Delta2 = self._decompose_Delta(self.Delta, self.K)

        for j in range(2*t+1):
            Theta0 = self._getTheta(j, self.Phi, Delta0*self.kappas[j+2])
            Theta1 = self._getTheta(j, self.Phi, Delta1*self.zetas[j+2])
            Theta2 = self._getTheta(j, self.Phi, Delta2*self.dzetas[j+2])
            Sigma += Theta0 + Theta1 + Theta1.T + Theta2

        # print("Sigma")
        # print(Sigma[:(t+1)*self.K, :(t+1)*self.K])
        # print("Delta")
        # print(self.Delta[:(t+1)*self.K, :(t+1)*self.K])
        
        return Sigma[:(t+1)*self.K, :(t+1)*self.K]
    
    def updateDelta(self, t, Delta):
        self.Delta[:(t+1)*self.K, :(t+1)*self.K] = Delta

    def updatePhi(self, t, phiRow):
        self.Phi[(t*self.K):(t+1)*self.K, :(t*self.K)] = phiRow
        return super().updatePhi(t, phiRow)


class RRISE(RISE):

    def __init__(self, K, niters, thetas, kappas, zetas, dzetas, separ):
        RISE.__init__(self, K, niters, thetas, kappas, zetas, dzetas)
        self.reg = 0
        self.singular_count = 0
    
    def getSigma(self, t):
        Sigma = super().getSigma(t)
        # Regularization
        D,Q = np.linalg.eigh(Sigma)
        singular_count = np.sum(D < 0)
        print("D at step {}".format(t))
        print(D)
        print("singular diff {} at step {}".format(singular_count - self.singular_count, t))
        self.singular_count = singular_count
        
        D = np.maximum(D, self.reg)
        Sigma = Q.dot(np.diag(D)).dot(np.transpose(Q))

        return Sigma 
    
class SRISE(RISE):

    def __init__(self, K, niters, thetas, kappas, zetas, dzetas, separ):
        RISE.__init__(self, K, niters, thetas, kappas, zetas, dzetas)
        # print("SRIES is used.")
        self.reg = 0
    
    def getSigma(self, t):
        Sigma = super().getSigma(t)
        # Regularization
        D,Q = np.linalg.eigh(Sigma)
        if D.min() < self.reg:
            print("D")
            print(D)
            raise RuntimeError("Hit stopping threshold at t ={}".format(t))

        return Sigma 



class GSE(RISE):
    def __init__(self, K, niters, thetas):
        kappas = np.array([0,0,1] + [0]*(niters*4-2))
        zetas = np.zeros((niters*4+1, K))
        zetas[0] = 1/thetas**2; zetas[1] = 1/thetas; zetas[2] = 1
        dzetas = np.zeros((niters*4+1, K))
        dzetas[0] = 3/thetas**2; dzetas[1] = 2/thetas; dzetas[2] = 1

        RISE.__init__(self, K, niters, thetas, kappas, zetas, dzetas)

    def getSigma(self, t):

        '''
        0 indicates the term doen's involve u_0
        1 indicates the term involving 1 u_0
        2 indicates the term involving 2 u_0
        we get the Sigma for f0, ... , fT
        '''
        # TODO check length for kappas, zetas and dzetas.
        # Monitor

        Sigma = np.zeros(self.shape)

        Delta0, Delta1, Delta2 = self._decompose_Delta(self.Delta, self.K)

        for j in range(2*t+1):
            Theta0 = self._getTheta(j, self.Phi, Delta0*self.kappas[j+2])
            Theta1 = self._getTheta(j, self.Phi, Delta1*self.zetas[j+2])
            Theta2 = self._getTheta(j, self.Phi, Delta2*self.dzetas[j+2])
            Sigma += Theta0 + Theta1 + Theta1.T + Theta2

        # print("Sigma")
        # print(Sigma[:(t+1)*self.K, :(t+1)*self.K])
        # print("Delta")
        # print(self.Delta[:(t+1)*self.K, :(t+1)*self.K])
        
        return Sigma[:(t+1)*self.K, :(t+1)*self.K]

from scipy import integrate
import scipy
def getDeltast(mus, Sigmas, mut, Sigmat):
    sdim = len(mus)
    tdim = len(mut)

    invSigmas = np.linalg.inv(Sigmas)
    invSigmasRoot = scipy.linalg.sqrtm(invSigmas)
    invSigmat = np.linalg.inv(Sigmat)
    invSigmatRoot = scipy.linalg.sqrtm(invSigmat)

    consts = mus.T @ invSigmas @ mus
    constt = mut.T @ invSigmat @ mut

    deforms = mus.T @ invSigmasRoot
    deformt = mut.T @ invSigmatRoot

    if tdim > sdim:
        def integrand(*args):
            assert len(args) == tdim
            x = np.array(args).reshape(tdim ,1)
            return np.tanh(consts + deforms @ x[:sdim]) * np.tanh(constt + deformt @ x) * np.exp(- (x*x).sum()/2)
    elif tdim == sdim:
        def integrand(*args):
            assert len(args) == tdim
            x = np.array(args).reshape(tdim ,1)
            return np.tanh(consts + deforms @ x)**2 * np.exp(-(x*x).sum()/2)
    
    support = [[-10,10]] * tdim
    res = integrate.nquad(integrand, support)[0] / np.sqrt(2 * np.pi)**tdim
    return res 



if __name__=="__main__":
    mus = np.array([1,2,3]).reshape((3,1))
    sigmas = np.identity(3)
    mut = np.array([1,2,3]).reshape((3,1))
    sigmat = np.identity(3)
    res = getDeltast(mus, sigmas, mut, sigmat)
    print(res)




        

    





